@@ -64,10 +64,11 @@ static void do_native_call(struct vm_method *method, void *target,
 			     unsigned long *args, union jvalue *result)
 {
 	unsigned long stack_size;
+	bool is_native = vm_method_is_vm_native(method);
 
 	push_args_on_stack(method, args, stack_size);
 
-	if (vm_method_is_vm_native(method)) {
+	if (is_native) {
 		if (vm_enter_vm_native(target, (void *) esp) < 0)
 			goto exit;
 	}
@@ -79,10 +80,10 @@ static void do_native_call(struct vm_method *method, void *target,
 		:
 		: [target] "m" (target),
 		  [result] "m" (result)
-		: "%edx", "%edi", "cc", "memory");
+		: "%eax", "%ecx", "%edx", "cc", "memory");
 
 exit:
-	if (vm_method_is_vm_native(method))
+	if (is_native)
 		vm_leave_vm_native();
 
 	restore_stack(stack_size);
@@ -92,10 +93,11 @@ static void do_native_call_xmm(struct vm_method *method, void *target,
 			     unsigned long *args, union jvalue *result)
 {
 	unsigned long stack_size;
+	bool is_native = vm_method_is_vm_native(method);
 
 	push_args_on_stack(method, args, stack_size);
 
-	if (vm_method_is_vm_native(method)) {
+	if (is_native) {
 		if (vm_enter_vm_native(target, (void *) esp) < 0)
 			goto exit;
 	}
@@ -109,7 +111,7 @@ static void do_native_call_xmm(struct vm_method *method, void *target,
 		: "%edx", "%edi", "cc", "memory");
 
 exit:
-	if (vm_method_is_vm_native(method))
+	if (is_native)
 		vm_leave_vm_native();
 
 	restore_stack(stack_size);
@@ -179,7 +181,8 @@ static unsigned long native_call_gp(struct vm_method *method,
 				    unsigned long *args)
 {
 	int i;
-	size_t sp = 0, r = 0;
+	size_t reg_count = 0;
+	size_t stack_count = 0;
 	struct vm_args_map *map = method->args_map;
 	unsigned long *stack, regs[6];
 	unsigned long result;
@@ -190,42 +193,45 @@ static unsigned long native_call_gp(struct vm_method *method,
 
 	for (i = 0; i < method->args_count; i++) {
 		if (map[i].reg == MACH_REG_UNASSIGNED)
-			stack[sp++] = args[i];
+			stack[stack_count++] = args[i];
 		else
-			regs[r++] = args[i];
+			regs[reg_count++] = args[i];
 
 		/* Skip duplicate slots. */
 		if (map[i].type == J_LONG || map[i].type == J_DOUBLE)
 			i++;
 	}
 
-	while (r < 6)
-		regs[r++] = 0;
+	while (reg_count < 6)
+		regs[reg_count++] = 0;
 
 	__asm__ volatile (
 		/* Copy stack arguments onto the stack. */
-		"movq %1, %%rsi \n"
-		"movq %2, %%rcx \n"
-		"shl $3, %%rcx \n"
-		"subq %%rcx, %%rsp \n"
-		"movq %%rsp, %%rdi \n"
-		"cld \n"
-		"rep movsq \n"
+		"	movq %[stack], %%rsi		\n"
+		"	movq %[stack_count], %%rcx	\n"
+		"	shl $3, %%rcx			\n"
+		"	subq %%rcx, %%rsp		\n"
+		"	movq %%rsp, %%rdi		\n"
+		"	cld				\n"
+		"	rep movsq			\n"
 
-		"movq %%rcx, %%r12 \n"
+		"	movq %%rcx, %%r12		\n"
 
 		/* Assign registers to register arguments. */
-		"movq 0x00(%4), %%rdi \n"
-		"movq 0x08(%4), %%rsi \n"
-		"movq 0x10(%4), %%rdx \n"
-		"movq 0x18(%4), %%rcx \n"
-		"movq 0x20(%4), %%r8 \n"
-		"movq 0x28(%4), %%r9 \n"
+		"	movq 0x00(%[regs]), %%rdi	\n"
+		"	movq 0x08(%[regs]), %%rsi	\n"
+		"	movq 0x10(%[regs]), %%rdx	\n"
+		"	movq 0x18(%[regs]), %%rcx	\n"
+		"	movq 0x20(%[regs]), %%r8	\n"
+		"	movq 0x28(%[regs]), %%r9	\n"
 
-		"call *%3 \n"
-		"addq %%r12, %%rsp \n"
+		"	call *%[target]			\n"
+		"	addq %%r12, %%rsp		\n"
 		: "=a" (result)
-		: "r" (stack), "r" (sp), "m" (target), "r" (regs)
+		: [target] "m" (target),
+		  [regs] "r" (regs),
+		  [stack] "r" (stack),
+		  [stack_count] "r" (stack_count)
 		: "rdi", "rsi", "rdx", "rcx", "r8", "r9", "r10", "r11", "r12", "cc", "memory"
 	);
 
@@ -234,15 +240,6 @@ static unsigned long native_call_gp(struct vm_method *method,
 	return result;
 }
 
-static unsigned long vm_native_call_gp(struct vm_method *method,
-				       const void *target,
-				       unsigned long *args)
-{
-	abort();
-
-	return 0;
-}
-
 /**
  * This calls a function with call arguments copied from @args
  * array. The array contains @args_count elements of machine word
@@ -289,53 +286,4 @@ void native_call(struct vm_method *method,
 	}
 }
 
-/**
- * This calls a VM native function with call arguments copied from
- * @args array. The array contains @args_count elements of machine
- * word size. The @target must be a pointer to a VM function. Call
- * result will be stored in @result.
- */
-void vm_native_call(struct vm_method *method,
-		    const void *target,
-		    unsigned long *args,
-		    union jvalue *result)
-{
-	switch (method->return_type.vm_type) {
-	case J_VOID:
-		vm_native_call_gp(method, target, args);
-		break;
-	case J_REFERENCE:
-		result->l = (jobject) vm_native_call_gp(method, target, args);
-		break;
-	case J_INT:
-		result->i = (jint) vm_native_call_gp(method, target, args);
-		break;
-	case J_CHAR:
-		result->c = (jchar) vm_native_call_gp(method, target, args);
-		break;
-	case J_BYTE:
-		result->b = (jbyte) vm_native_call_gp(method, target, args);
-		break;
-	case J_SHORT:
-		result->s = (jshort) vm_native_call_gp(method, target, args);
-		break;
-	case J_BOOLEAN:
-		result->z = (jboolean) vm_native_call_gp(method, target, args);
-		break;
-	case J_LONG:
-		result->j = (jlong) vm_native_call_gp(method, target, args);
-		break;
-	case J_DOUBLE:
-	case J_FLOAT:
-		error("not implemented");
-		break;
-	case J_RETURN_ADDRESS:
-	case VM_TYPE_MAX:
-		die("unexpected type");
-	}
-}
-
-#warning NOT IMPLEMENTED
-
 #endif /* CONFIG_X86_32 */
-
