@@ -25,37 +25,70 @@
  * Please refer to the file LICENSE for details.
  */
 
+#include "vm/class.h"
+
+#include "cafebabe/enclosing_method_attribute.h"
+#include "cafebabe/inner_classes_attribute.h"
+#include "cafebabe/annotations_attribute.h"
+#include "cafebabe/constant_pool.h"
+#include "cafebabe/method_info.h"
+#include "cafebabe/field_info.h"
+#include "cafebabe/stream.h"
+#include "cafebabe/class.h"
+
+#include "jit/exception.h"
+#include "jit/compiler.h"
+#include "jit/vtable.h"
+#include "jit/cu-mapping.h"
+
+#include "vm/fault-inject.h"
+#include "vm/classloader.h"
+#include "vm/annotation.h"
+#include "vm/gc.h"
+#include "vm/preload.h"
+#include "vm/errors.h"
+#include "vm/itable.h"
+#include "vm/method.h"
+#include "vm/object.h"
+#include "vm/stdlib.h"
+#include "vm/thread.h"
+#include "vm/field.h"
+#include "vm/die.h"
+#include "vm/vm.h"
+#include "vm/trace.h"
+
+#include "lib/string.h"
+#include "lib/array.h"
+
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
 #include <stdio.h>
 #include <ctype.h>
 
-#include "cafebabe/class.h"
-#include "cafebabe/constant_pool.h"
-#include "cafebabe/field_info.h"
-#include "cafebabe/method_info.h"
-#include "cafebabe/stream.h"
+bool opt_trace_vtable;
 
-#include "lib/array.h"
-#include "lib/string.h"
+static void
+trace_vtable(struct vm_class *vmc)
+{
+	struct vtable *vtable = &(vmc->vtable);
+	struct compilation_unit *cu;
+	struct vm_method *vmm;
 
-#include "vm/class.h"
-#include "vm/classloader.h"
-#include "vm/die.h"
-#include "vm/fault-inject.h"
-#include "vm/field.h"
-#include "vm/preload.h"
-#include "vm/itable.h"
-#include "vm/method.h"
-#include "vm/object.h"
-#include "vm/stdlib.h"
-#include "vm/thread.h"
-#include "vm/vm.h"
+	trace_printf("trace vtable: %s\n", vmc->name);
+
+	for (uint16_t i = 0; i < vmc->vtable_size; ++i) {
+		cu = jit_lookup_cu((long unsigned int)vtable->native_ptr[i]);
+		if (cu) {
+			vmm = cu->method;
+			trace_printf("  vtable[%d] = %s.%s%s\n", i, vmm->class->name,
+				     vmm->name, vmm->type);
+		}
+	}
+
+	trace_flush();
+}
 
-#include "jit/exception.h"
-#include "jit/compiler.h"
-#include "jit/vtable.h"
 
 static void
 setup_vtable(struct vm_class *vmc)
@@ -72,6 +105,7 @@ setup_vtable(struct vm_class *vmc)
 		super_vtable = &super->vtable;
 	} else {
 		super_vtable_size = 0;
+		super_vtable = NULL;
 	}
 
 	vtable_size = 0;
@@ -109,24 +143,49 @@ setup_vtable(struct vm_class *vmc)
 				    vmm->virtual_index,
 				    vm_method_call_ptr(vmm));
 	}
+
+	if (opt_trace_vtable)
+		trace_vtable(vmc);
+}
+
+/*
+ * Set the .object member of struct vm_class to point to the object
+ * of type java.lang.Class for this class.
+ */
+int vm_class_setup_object(struct vm_class *vmc)
+{
+	vmc->object = vm_object_alloc(vm_java_lang_Class);
+	if (!vmc->object) {
+		throw_oom_error();
+		return -1;
+	}
+
+	field_set_object(vmc->object, vm_java_lang_Class_vmdata,
+			 (struct vm_object *)vmc);
+	return 0;
 }
 
 static int vm_class_link_common(struct vm_class *vmc)
 {
 	int err;
 
+	compile_lock_init(&vmc->cl, true);
+
 	err = pthread_mutex_init(&vmc->mutex, NULL);
 	if (err)
 		return -err;
 
-	err = vm_monitor_init(&vmc->monitor);
-	if (err)
-		return -err;
+	if (preload_finished)
+		return vm_class_setup_object(vmc);
 
-	vmc->object = NULL;
-	vmc->classloader = NULL;
+	/* If classes and fields were not preloaded yet then we can't
+	 * allocate object for this class. Set .object to a dummy object
+	 * for class synchronization to work. This will be replaced
+	 * after preloading is finished. */
+	static struct vm_object dummy_object;
+	vmc->object = &dummy_object;
 
-	return 0;
+	return vm_preload_add_class_fixup(vmc);
 }
 
 /*
@@ -221,29 +280,34 @@ static int compare_method_signatures(const void *a, const void *b)
 	return 0;
 }
 
+static void free_buckets(int rows, int cols, struct field_bucket field_buckets[rows][cols])
+{
+	for (int i = 0; i < cols; ++i) {
+		for (int j = 0; j < rows; ++j) {
+			struct field_bucket *bucket = &field_buckets[j][i];
+
+			vm_free(bucket->fields);
+		}
+	}
+}
+
 int vm_class_link(struct vm_class *vmc, const struct cafebabe_class *class)
 {
+	const struct cafebabe_constant_info_class *constant_class;
+	const struct cafebabe_constant_info_utf8 *name;
+	unsigned int nr_inner_classes;
+
 	vmc->class = class;
 	vmc->kind = VM_CLASS_KIND_REGULAR;
 
 	if (vm_class_link_common(vmc))
 		return -1;
 
-	const struct cafebabe_constant_info_class *constant_class;
-	if (cafebabe_class_constant_get_class(class,
-		class->this_class, &constant_class))
-	{
-		NOT_IMPLEMENTED;
+	if (cafebabe_class_constant_get_class(class, class->this_class, &constant_class))
 		return -1;
-	}
 
-	const struct cafebabe_constant_info_utf8 *name;
-	if (cafebabe_class_constant_get_utf8(class,
-		constant_class->name_index, &name))
-	{
-		NOT_IMPLEMENTED;
+	if (cafebabe_class_constant_get_utf8(class, constant_class->name_index, &name))
 		return -1;
-	}
 
 	vmc->name = strndup((char *) name->bytes, name->length);
 
@@ -253,93 +317,67 @@ int vm_class_link(struct vm_class *vmc, const struct cafebabe_class *class)
 
 	if (class->super_class) {
 		const struct cafebabe_constant_info_class *constant_super;
-		if (cafebabe_class_constant_get_class(class,
-			class->super_class, &constant_super))
-		{
-			NOT_IMPLEMENTED;
-			return -1;
-		}
-
 		const struct cafebabe_constant_info_utf8 *super_name;
-		if (cafebabe_class_constant_get_utf8(class,
-			constant_super->name_index, &super_name))
-		{
-			NOT_IMPLEMENTED;
-			return -1;
-		}
 
-		char *super_name_str = strndup((char *) super_name->bytes,
-			super_name->length);
+		if (cafebabe_class_constant_get_class(class, class->super_class, &constant_super))
+			goto error_free_name;
+
+		if (cafebabe_class_constant_get_utf8(class, constant_super->name_index, &super_name))
+			goto error_free_name;
+
+		char *super_name_str = strndup((char *) super_name->bytes, super_name->length);
 
 		/* XXX: Circularity check */
-		vmc->super = classloader_load(NULL, super_name_str);
-		if (!vmc->super) {
-			NOT_IMPLEMENTED;
-			return -1;
-		}
-
+		vmc->super = classloader_load(vmc->classloader, super_name_str);
 		free(super_name_str);
+
+		if (!vmc->super)
+			goto error_free_name;
 	} else {
-		if (!strcmp(vmc->name, "java.lang.Object")) {
-			NOT_IMPLEMENTED;
-			return -1;
-		}
+		if (!strcmp(vmc->name, "java.lang.Object"))
+			goto error_free_name;
 
 		vmc->super = NULL;
 	}
 
 	vmc->nr_interfaces = class->interfaces_count;
-	vmc->interfaces
-		= malloc(sizeof(*vmc->interfaces) * vmc->nr_interfaces);
-	if (!vmc->interfaces) {
-		NOT_IMPLEMENTED;
-		return -1;
-	}
+	vmc->interfaces = malloc(sizeof(*vmc->interfaces) * vmc->nr_interfaces);
+	if (!vmc->interfaces)
+		goto error_free_name;
 
 	for (unsigned int i = 0; i < class->interfaces_count; ++i) {
+		const struct cafebabe_constant_info_class *interface;
+		const struct cafebabe_constant_info_utf8 *name;
 		uint16_t idx = class->interfaces[i];
 
-		const struct cafebabe_constant_info_class *interface;
-		if (cafebabe_class_constant_get_class(class, idx, &interface)) {
-			NOT_IMPLEMENTED;
-			return -1;
-		}
+		if (cafebabe_class_constant_get_class(class, idx, &interface))
+			goto error_free_interfaces;
 
-		const struct cafebabe_constant_info_utf8 *name;
-		if (cafebabe_class_constant_get_utf8(class,
-			interface->name_index, &name))
-		{
-			NOT_IMPLEMENTED;
-			return -1;
-		}
+		if (cafebabe_class_constant_get_utf8(class, interface->name_index, &name))
+			goto error_free_interfaces;
 
 		char *c_name = strndup((char *) name->bytes, name->length);
-		if (!c_name) {
-			NOT_IMPLEMENTED;
-			return -1;
-		}
+		if (!c_name)
+			goto error_free_interfaces;
 
-		struct vm_class *vmi = classloader_load(NULL, c_name);
+		struct vm_class *vmi = classloader_load(vmc->classloader, c_name);
 		free(c_name);
 		if (!vmi)
-			return -1;
+			goto error_free_interfaces;
 
 		vmc->interfaces[i] = vmi;
 	}
 
-	vmc->fields = malloc(sizeof(*vmc->fields) * class->fields_count);
-	if (!vmc->fields) {
-		NOT_IMPLEMENTED;
-		return -1;
-	}
+	vmc->nr_fields = class->fields_count;
+	vmc->fields = vm_alloc(sizeof(*vmc->fields) * class->fields_count);
+	if (!vmc->fields)
+		goto error_free_interfaces;
 
-	for (uint16_t i = 0; i < class->fields_count; ++i) {
+	for (uint16_t i = 0; i < vmc->nr_fields; ++i) {
 		struct vm_field *vmf = &vmc->fields[i];
 
-		if (vm_field_init(vmf, vmc, i)) {
-			NOT_IMPLEMENTED;
-			return -1;
-		}
+		if (vm_field_init(vmf, vmc, i))
+			goto error_free_fields;
 	}
 
 	if (vmc->super) {
@@ -360,7 +398,7 @@ int vm_class_link(struct vm_class *vmc, const struct cafebabe_class *class)
 	for (uint16_t i = 0; i < class->fields_count; ++i) {
 		struct vm_field *vmf = &vmc->fields[i];
 		unsigned int stat = vm_field_is_static(vmf) ? 0 : 1;
-		enum vm_type type = str_to_type(vmf->type);
+		enum vm_type type = vmf->type_info.vm_type;
 		struct field_bucket *bucket = &field_buckets[stat][type];
 
 		++bucket->nr;
@@ -371,7 +409,7 @@ int vm_class_link(struct vm_class *vmc, const struct cafebabe_class *class)
 		for (unsigned int j = 0; j < 2; ++j) {
 			struct field_bucket *bucket = &field_buckets[j][i];
 
-			bucket->fields = malloc(bucket->nr * sizeof(*bucket->fields));
+			bucket->fields = vm_alloc(bucket->nr * sizeof(*bucket->fields));
 			bucket->nr = 0;
 		}
 	}
@@ -380,7 +418,7 @@ int vm_class_link(struct vm_class *vmc, const struct cafebabe_class *class)
 	for (uint16_t i = 0; i < class->fields_count; ++i) {
 		struct vm_field *vmf = &vmc->fields[i];
 		unsigned int stat = vm_field_is_static(vmf) ? 0 : 1;
-		enum vm_type type = str_to_type(vmf->type);
+		enum vm_type type = vmf->type_info.vm_type;
 		struct field_bucket *bucket = &field_buckets[stat][type];
 
 		bucket->fields[bucket->nr++] = vmf;
@@ -391,16 +429,16 @@ int vm_class_link(struct vm_class *vmc, const struct cafebabe_class *class)
 	buckets_order_fields(field_buckets[1], &tmp, &vmc->object_size);
 
 	/* XXX: only static fields, right size, etc. */
-	vmc->static_values = zalloc(vmc->static_size);
+	vmc->static_values = vm_zalloc(vmc->static_size);
+	if (!vmc->static_values)
+		goto error_free_buckets;
 
-	for (uint16_t i = 0; i < class->fields_count; ++i) {
+	for (uint16_t i = 0; i < vmc->nr_fields; ++i) {
 		struct vm_field *vmf = &vmc->fields[i];
 
 		if (vm_field_is_static(vmf)) {
-			if (vm_field_init_static(vmf)) {
-				NOT_IMPLEMENTED;
-				return -1;
-			}
+			if (vm_field_init_static(vmf))
+				goto error_free_static_values;
 		} else {
 			vm_field_init_nonstatic(vmf);
 		}
@@ -423,13 +461,11 @@ int vm_class_link(struct vm_class *vmc, const struct cafebabe_class *class)
 
 		for (unsigned int j = 0; j < vmi->nr_methods; ++j) {
 			struct vm_method *vmm = &vmi->methods[j];
+			int err;
 
-			int err = insert_interface_method(vmc,
-				&extra_methods, vmm);
-			if (err) {
-				NOT_IMPLEMENTED;
-				return -1;
-			}
+			err = insert_interface_method(vmc, &extra_methods, vmm);
+			if (err)
+				goto error_free_static_values;
 		}
 	}
 
@@ -442,30 +478,22 @@ int vm_class_link(struct vm_class *vmc, const struct cafebabe_class *class)
 
 	vmc->nr_methods = class->methods_count + extra_methods.size;
 
-	vmc->methods = malloc(sizeof(*vmc->methods) * vmc->nr_methods);
-	if (!vmc->methods) {
-		NOT_IMPLEMENTED;
-		return -1;
-	}
+	vmc->methods = vm_alloc(sizeof(*vmc->methods) * vmc->nr_methods);
+	if (!vmc->methods)
+		goto error_free_static_values;
 
 	for (uint16_t i = 0; i < class->methods_count; ++i) {
 		struct vm_method *vmm = &vmc->methods[i];
 
-		if (vm_method_init(vmm, vmc, i)) {
-			NOT_IMPLEMENTED;
-			return -1;
-		}
+		if (vm_method_init(vmm, vmc, i))
+			goto error_free_methods;
 	}
 
 	for (unsigned int i = 0; i < extra_methods.size; ++i) {
 		struct vm_method *vmm = &vmc->methods[class->methods_count + i];
 
-		if (vm_method_init_from_interface(vmm, vmc,
-			class->methods_count + i, extra_methods.ptr[i]))
-		{
-			NOT_IMPLEMENTED;
-			return -1;
-		}
+		if (vm_method_init_from_interface(vmm, vmc, class->methods_count + i, extra_methods.ptr[i]))
+			goto error_free_methods;
 	}
 
 	for (uint16_t i = 0; i < vmc->nr_methods; ++i) {
@@ -473,10 +501,8 @@ int vm_class_link(struct vm_class *vmc, const struct cafebabe_class *class)
 
 		vmm->itable_index = itable_hash(vmm);
 
-		if (vm_method_prepare_jit(vmm)) {
-			NOT_IMPLEMENTED;
-			return -1;
-		}
+		if (vm_method_prepare_jit(vmm))
+			goto error_free_methods;
 	}
 
 	array_destroy(&extra_methods);
@@ -490,8 +516,87 @@ int vm_class_link(struct vm_class *vmc, const struct cafebabe_class *class)
 
 	INIT_LIST_HEAD(&vmc->static_fixup_site_list);
 
+	struct cafebabe_inner_classes_attribute inner_classes_attribute;
+
+	memset(&inner_classes_attribute, 0, sizeof(inner_classes_attribute));
+	if (cafebabe_read_inner_classes_attribute(class, &class->attributes, &inner_classes_attribute))
+		return -1;
+
+	nr_inner_classes = 0;
+	for (unsigned int i = 0; i < inner_classes_attribute.number_of_classes; i++) {
+		struct cafebabe_inner_class *inner = &inner_classes_attribute.inner_classes[i];
+
+		if (class->this_class == inner->outer_class_info_index)
+			nr_inner_classes++;
+	}
+
+	vmc->inner_classes = vm_alloc(sizeof(*vmc->inner_classes) * nr_inner_classes);
+	if (!vmc->inner_classes)
+		goto error_free_methods;
+
+	for (unsigned int i = 0; i < inner_classes_attribute.number_of_classes; i++) {
+		struct cafebabe_inner_class *inner = &inner_classes_attribute.inner_classes[i];
+
+		if (class->this_class == inner->inner_class_info_index) {
+			vmc->declaring_class = vmc->enclosing_class = vm_class_resolve_class(vmc, inner->outer_class_info_index);
+
+			vmc->inner_class_access_flags	= inner->inner_class_access_flags;
+		} else if (class->this_class == inner->outer_class_info_index) {
+			vmc->inner_classes[vmc->nr_inner_classes++] = inner->inner_class_info_index;
+		}
+	}
+
+	struct cafebabe_annotations_attribute annotations_attribute;
+
+	if (cafebabe_read_annotations_attribute(class, &class->attributes, &annotations_attribute))
+		goto error_free_inner_classes;
+
+	vmc->annotations = vm_alloc(sizeof(struct vm_annotation *) * annotations_attribute.num_annotations);
+	if (!vmc->annotations)
+		goto error_free_inner_classes;
+
+	for (unsigned int i = 0; i < annotations_attribute.num_annotations; i++) {
+		struct cafebabe_annotation *annotation = &annotations_attribute.annotations[i];
+		struct vm_annotation *vma;
+
+		vma = vm_annotation_parse(vmc, annotation);
+		if (!vma)
+			goto error_free_annotations;
+
+		vmc->annotations[vmc->nr_annotations++] = vma;
+	}
+
+	cafebabe_annotations_attribute_deinit(&annotations_attribute);
+
+	if (!cafebabe_read_enclosing_method_attribute(class, &class->attributes, &vmc->enclosing_method_attribute))
+		vmc->enclosing_class = vm_class_resolve_class(vmc, vmc->enclosing_method_attribute.class_index);
+
 	vmc->state = VM_CLASS_LINKED;
 	return 0;
+
+error_free_annotations:
+	for (unsigned int i = 0; i < vmc->nr_annotations; i++) {
+		struct vm_annotation *vma = vmc->annotations[i];
+
+		vm_annotation_free(vma);
+	}
+	vm_free(vmc->annotations);
+error_free_methods:
+	vm_free(vmc->methods);
+error_free_inner_classes:
+	vm_free(vmc->inner_classes);
+error_free_static_values:
+	vm_free(vmc->static_values);
+error_free_buckets:
+	free_buckets(2, VM_TYPE_MAX, field_buckets);
+error_free_fields:
+	vm_free(vmc->fields);
+error_free_interfaces:
+	free(vmc->interfaces);
+error_free_name:
+	free(vmc->name);
+
+	return -1;
 }
 
 int vm_class_link_primitive_class(struct vm_class *vmc, const char *class_name)
@@ -581,61 +686,31 @@ static bool vm_class_check_class_init_fault(struct vm_class *vmc,
 
 int vm_class_ensure_object(struct vm_class *vmc)
 {
-	vm_monitor_lock(&vmc->monitor);
-
-	if (vmc->object)
-		goto out;
-
-	/*
-	 * Set the .object member of struct vm_class to point to
-	 * the object (of type java.lang.Class) for this class.
-	 */
-	vmc->object = vm_object_alloc(vm_java_lang_Class);
-	if (!vmc->object) {
-		NOT_IMPLEMENTED;
-		vm_monitor_unlock(&vmc->monitor);
-		return -1;
-	}
-
-	field_set_object(vmc->object, vm_java_lang_Class_vmdata,
-		(struct vm_object *)vmc);
-
- out:
-	vm_monitor_unlock(&vmc->monitor);
+	/* Before preload is finished .object points to a common
+	 * dummy object. */
+	assert(preload_finished);
 	return 0;
 }
 
 int vm_class_init(struct vm_class *vmc)
 {
 	struct vm_object *exception;
+	enum compile_lock_status status;
 
-	vm_monitor_lock(&vmc->monitor);
+	status = compile_lock_enter(&vmc->cl);
+	if (status == STATUS_COMPILED_OK || status == STATUS_REENTER)
+		return 0;
 
-	if (vmc->state == VM_CLASS_INITIALIZING) {
-		/* XXX: we need to break recursion. */
-		if (vmc->initializing_thread == vm_thread_self())
-			goto out_unlock;
-
-		while (vmc->state == VM_CLASS_INITIALIZING)
-			vm_monitor_wait(&vmc->monitor);
-	}
-
-	if (vmc->state == VM_CLASS_INITIALIZED)
-		goto out_unlock;
-
-	if (vmc->state == VM_CLASS_ERRONEOUS) {
+	if (status == STATUS_COMPILED_ERRONOUS) {
 		signal_new_exception(vm_java_lang_NoClassDefFoundError,
 				     vmc->name);
-		vm_monitor_unlock(&vmc->monitor);
 		return -1;
 	}
 
+	vm_object_lock(vmc->object);
 	assert(vmc->state == VM_CLASS_LINKED);
-
 	vmc->state = VM_CLASS_INITIALIZING;
-	vmc->initializing_thread = vm_thread_self();
-
-	vm_monitor_unlock(&vmc->monitor);
+	vm_object_unlock(vmc->object);
 
 	/* Fault injection, for testing purposes */
 	if (vm_fault_enabled(VM_FAULT_CLASS_INIT)) {
@@ -676,18 +751,16 @@ int vm_class_init(struct vm_class *vmc)
 		}
 	}
 
-	vm_monitor_lock(&vmc->monitor);
+	vm_object_lock(vmc->object);
 	vmc->state = VM_CLASS_INITIALIZED;
-	vm_monitor_notify_all(&vmc->monitor);
-	vm_monitor_unlock(&vmc->monitor);
+	vm_object_notify_all(vmc->object);
+	vm_object_unlock(vmc->object);
 
-	return 0;
-
- out_unlock:
-	vm_monitor_unlock(&vmc->monitor);
+	compile_lock_leave(&vmc->cl, STATUS_COMPILED_OK);
 	return 0;
 
  error:
+	compile_lock_leave(&vmc->cl, STATUS_COMPILED_ERRONOUS);
 	exception = exception_occurred();
 
 	if (!vm_object_is_instance_of(exception, vm_java_lang_Error)) {
@@ -697,10 +770,10 @@ int vm_class_init(struct vm_class *vmc)
 			vmc->name);
 	}
 
-	vm_monitor_lock(&vmc->monitor);
+	vm_object_lock(vmc->object);
 	vmc->state = VM_CLASS_ERRONEOUS;
-	vm_monitor_notify_all(&vmc->monitor);
-	vm_monitor_unlock(&vmc->monitor);
+	vm_object_notify_all(vmc->object);
+	vm_object_unlock(vmc->object);
 
 	return -1;
 }
@@ -708,12 +781,9 @@ int vm_class_init(struct vm_class *vmc)
 struct vm_class *vm_class_resolve_class(const struct vm_class *vmc, uint16_t i)
 {
 	const struct cafebabe_constant_info_class *constant_class;
-	if (cafebabe_class_constant_get_class(vmc->class,
-		i, &constant_class))
-	{
-		NOT_IMPLEMENTED;
+
+	if (cafebabe_class_constant_get_class(vmc->class, i, &constant_class))
 		return NULL;
-	}
 
 	const struct cafebabe_constant_info_utf8 *class_name;
 	if (cafebabe_class_constant_get_utf8(vmc->class,
@@ -733,7 +803,7 @@ struct vm_class *vm_class_resolve_class(const struct vm_class *vmc, uint16_t i)
 	struct vm_class *class
 		= classloader_load(vmc->classloader, class_name_str);
 	if (!class) {
-		NOT_IMPLEMENTED;
+		warn("failed to load class %s", class_name_str);
 		return NULL;
 	}
 
@@ -853,33 +923,17 @@ vm_class_resolve_field_recursive(const struct vm_class *vmc, uint16_t i)
 	return result;
 }
 
-int vm_class_resolve_method(const struct vm_class *vmc, uint16_t i,
-	struct vm_class **r_vmc, char **r_name, char **r_type)
+static int vm_class_resolve_name_and_type(const struct vm_class *vmc, uint16_t index, char **r_name, char **r_type)
 {
-	const struct cafebabe_constant_info_method_ref *method;
-	if (cafebabe_class_constant_get_method_ref(vmc->class, i, &method)) {
-		NOT_IMPLEMENTED;
-		return -1;
-	}
-
-	struct vm_class *class = vm_class_resolve_class(vmc,
-		method->class_index);
-	if (!class) {
-		NOT_IMPLEMENTED;
-		return -1;
-	}
-
 	const struct cafebabe_constant_info_name_and_type *name_and_type;
-	if (cafebabe_class_constant_get_name_and_type(vmc->class,
-		method->name_and_type_index, &name_and_type))
-	{
+
+	if (cafebabe_class_constant_get_name_and_type(vmc->class, index, &name_and_type)) {
 		NOT_IMPLEMENTED;
 		return -1;
 	}
 
 	const struct cafebabe_constant_info_utf8 *name;
-	if (cafebabe_class_constant_get_utf8(vmc->class,
-		name_and_type->name_index, &name))
+	if (cafebabe_class_constant_get_utf8(vmc->class, name_and_type->name_index, &name))
 	{
 		NOT_IMPLEMENTED;
 		return -1;
@@ -905,9 +959,32 @@ int vm_class_resolve_method(const struct vm_class *vmc, uint16_t i,
 		return -1;
 	}
 
-	*r_vmc = class;
 	*r_name = name_str;
 	*r_type = type_str;
+
+	return 0;
+}
+
+int vm_class_resolve_method(const struct vm_class *vmc, uint16_t i,
+	struct vm_class **r_vmc, char **r_name, char **r_type)
+{
+	const struct cafebabe_constant_info_method_ref *method;
+	if (cafebabe_class_constant_get_method_ref(vmc->class, i, &method)) {
+		NOT_IMPLEMENTED;
+		return -1;
+	}
+
+	struct vm_class *class = vm_class_resolve_class(vmc, method->class_index);
+	if (!class) {
+		NOT_IMPLEMENTED;
+		return -1;
+	}
+
+	if (vm_class_resolve_name_and_type(vmc, method->name_and_type_index, r_name, r_type))
+		return -1;
+
+	*r_vmc = class;
+
 	return 0;
 }
 
@@ -1018,13 +1095,56 @@ int vm_class_resolve_interface_method(const struct vm_class *vmc, uint16_t i,
 	return 0;
 }
 
-struct vm_method *
-vm_class_resolve_method_recursive(const struct vm_class *vmc, uint16_t i)
+static struct vm_method *
+missing_method(const struct vm_class *vmc, char *name, char *type, uint16_t access_flags)
 {
+	struct cafebabe_method_info *method;
+	struct vm_method *vmm;
+
+	method		= vm_alloc(sizeof *method);
+	if (!method)
+		return NULL;
+
+	method->access_flags	= access_flags;
+
+	vmm		= vm_alloc(sizeof *vmm);
+	if (!vmm)
+		goto error_free_method;
+
+	vmm->name	= strdup(name);
+	if (!vmm->name)
+		goto error_free_vmm;
+
+	vmm->type	= strdup(type);
+	if (!vmm->type)
+		goto error_free_name;
+
+	vmm->method	= method;
+
+	if (vm_method_do_init(vmm))
+		goto error_free_type;
+
+	return vmm;
+
+error_free_type:
+	free(vmm->type);
+error_free_name:
+	free(vmm->name);
+error_free_vmm:
+	free(vmm);
+error_free_method:
+	free(method);
+
+	return NULL;
+}
+
+struct vm_method *
+vm_class_resolve_method_recursive(const struct vm_class *vmc, uint16_t i, uint16_t access_flags)
+{
+	struct vm_method *result;
 	struct vm_class *class;
 	char *name;
 	char *type;
-	struct vm_method *result;
 
 	if (vm_class_resolve_method(vmc, i, &class, &name, &type)) {
 		NOT_IMPLEMENTED;
@@ -1032,9 +1152,12 @@ vm_class_resolve_method_recursive(const struct vm_class *vmc, uint16_t i)
 	}
 
 	result = vm_class_get_method_recursive(class, name, type);
+	if (!result)
+		result = missing_method(class, name, type, access_flags);
 
 	free(name);
 	free(type);
+
 	return result;
 }
 
@@ -1069,8 +1192,28 @@ static bool is_numeric(const char *s)
 	return true;
 }
 
+struct vm_method *vm_class_get_enclosing_method(const struct vm_class *vmc)
+{
+	struct vm_method *result;
+	char *name;
+	char *type;
+
+	if (!vmc->enclosing_class)
+		return NULL;
+
+	if (vm_class_resolve_name_and_type(vmc, vmc->enclosing_method_attribute.method_index, &name, &type))
+		return NULL;
+
+	result = vm_class_get_method_recursive(vmc->enclosing_class, name, type);
+
+	free(name);
+	free(type);
+
+	return result;
+}
+
 /* See Section 15.9.5 ("Anonymous Class Declarations") of the JLS for details.  */
-bool vm_class_is_anonymous(struct vm_class *vmc)
+bool vm_class_is_anonymous(const struct vm_class *vmc)
 {
 	if (!vm_class_is_regular_class(vmc))
 		return false;
@@ -1089,7 +1232,7 @@ bool vm_class_is_anonymous(struct vm_class *vmc)
 	return is_numeric(separator + 1);
 }
 
-/* Reference: http://java.sun.com/j2se/1.5.0/docs/api/java/lang/Class.html#isAssignableFrom(java.lang.Class) */
+/* Reference: http://download.oracle.com/javase/1.5.0/docs/api/java/lang/Class.html#isAssignableFrom(java.lang.Class) */
 bool vm_class_is_assignable_from(const struct vm_class *vmc, const struct vm_class *from)
 {
 	if (vmc == from)
@@ -1162,12 +1305,44 @@ struct vm_class *vm_class_get_array_class(struct vm_class *element_class)
 	struct vm_class *result;
 	char *name;
 
-	if (!asprintf(&name, "[%s", element_class->name)) {
-		signal_new_exception(vm_java_lang_OutOfMemoryError, NULL);
-		return NULL;
-	}
+	if (!asprintf(&name, "[L%s;", element_class->name))
+		return throw_oom_error();
 
 	result = classloader_load(element_class->classloader, name);
 	free(name);
 	return result;
 }
+
+struct vm_class *
+vm_class_define(struct vm_object *classloader, const char *name,
+		uint8_t *data, unsigned long len)
+{
+	struct cafebabe_stream stream;
+	struct cafebabe_class *class;
+	struct vm_class *result = NULL;
+
+	cafebabe_stream_open_buffer(&stream, data, len);
+
+	class = malloc(sizeof *class);
+	if (cafebabe_class_init(class, &stream)) {
+		signal_new_exception(vm_java_lang_ClassFormatError, NULL);
+		goto out_stream;
+	}
+
+	cafebabe_stream_close_buffer(&stream);
+
+	result = vm_alloc(sizeof *result);
+	if (!result)
+		return throw_oom_error();
+
+	result->classloader = classloader;
+
+	if (vm_class_link(result, class))
+		return NULL;
+
+	return result;
+
+out_stream:
+	cafebabe_stream_close_buffer(&stream);
+	return NULL;
+}
